# SPDX-FileCopyrightText: 2021 The octopy-energy Authors
#
# SPDX-License-Identifier: Apache-2.0
"""GraphQL client for Octopus Energy APIs."""

import datetime
from typing import Mapping, Optional, Tuple

import gql
import gql.transport.aiohttp

from . import Consumption, Tariff

_OCTOPUS_ENERGY_API_URL = "https://api.octopus.energy/v1/graphql/"


class ClientSession:

    _apikey: str
    _api_url: str

    _headers: Optional[Mapping[str, str]] = {}
    _authenticated_client: Optional[gql.Client] = None

    def __init__(self, apikey: str, api_url: str = _OCTOPUS_ENERGY_API_URL) -> None:
        self._apikey = apikey
        self._api_url = api_url

    def _get_transport(self) -> gql.transport.aiohttp.AIOHTTPTransport:
        return gql.transport.aiohttp.AIOHTTPTransport(
            self._api_url, headers=self._headers
        )

    def _get_client(self) -> gql.Client:
        return gql.Client(
            transport=self._get_transport(), fetch_schema_from_transport=False
        )

    async def __aenter__(self) -> "ClientSession":
        self._headers = {"Authorization": await self.obtain_kraken_token()}
        self._authenticated_client = self._get_client()
        await self._authenticated_client.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc, tb) -> None:
        if self._authenticated_client is not None:
            await self._authenticated_client.__aexit__(exc_type, exc, tb)
        self._authenticated_client = None

    async def obtain_kraken_token(self) -> str:
        # obtain_kraken_token is always called with a new session, since it does not
        # require authentication.
        async with self._get_client() as session:
            result = await session.execute(
                _MUTATION_OBTAIN_KRAKEN_TOKEN, variable_values={"apikey": self._apikey}
            )
            return result["obtainKrakenToken"]["token"]

    async def query_account_active_electricity_tariffs(
        self, account_number: str
    ) -> Mapping[str, Tariff]:
        assert self._authenticated_client is not None
        result = await self._authenticated_client.session.execute(
            _QUERY_ACCOUNT_ACTIVE_TARIFFS,
            variable_values={"accountNumber": account_number},
        )
        return {
            agreement["meterPoint"]["mpan"]: Tariff.from_graphql(agreement["tariff"])
            for agreement in result["account"]["electricityAgreements"]
        }

    async def query_account_current_electricity_consumption_and_tariff(
        self, account_number: str
    ) -> Mapping[str, Tuple[Mapping[str, Consumption], Tariff]]:
        assert self._authenticated_client is not None
        # We need to define a date to start looking up the current consumption.
        # Since we're looking for a gauge rather than a measurement, we aggregate by quarter
        # and that means we can just start from four months ago, asking for the last.
        start_date = datetime.datetime.utcnow().astimezone(
            datetime.timezone.utc
        ) - datetime.timedelta(weeks=16)

        raw_result = await self._authenticated_client.session.execute(
            _QUERY_ACCOUNT_CURRENT_CONSUMPTION_AND_TARIFF,
            variable_values={
                "accountNumber": account_number,
                "startAt": start_date.isoformat(),
            },
        )

        result = {}
        for property in raw_result["properties"]:
            for meter_point in property["electricityMeterPoints"]:
                # There should only be a single active tariff.
                assert len(meter_point["agreements"]) == 1

                tariff = Tariff.from_graphql(meter_point["agreements"][0])
                consumptions = {}
                for meter in meter_point["meters"]:
                    # There should only be one consumption value.
                    assert len(meter["consumption"]["edges"]) == 1
                    consumptions[meter["serialNumber"]] = Consumption.from_graphql(
                        meter["consumption"]["edges"][0]
                    )
                result[meter_point["mpan"]] = (consumptions, tariff)

        return result


_MUTATION_OBTAIN_KRAKEN_TOKEN = gql.gql(
    """
        mutation krakenTokenAuthentication($apikey: String!) {
            obtainKrakenToken(input:{APIKey:$apikey}) {
                token
            }
        }
    """
)

_QUERY_ACCOUNT_ACTIVE_TARIFFS = gql.gql(
    """
    query accountActiveTariff($accountNumber: String!) {
        account(accountNumber: $accountNumber) {
            electricityAgreements(active:true) {
                meterPoint{
                    mpan
                }
                tariff {
                    ... on StandardTariff {
                        displayName
                        unitRate
                    }
                }
            }
        }
    }
    """
)

_QUERY_ACCOUNT_CURRENT_CONSUMPTION_AND_TARIFF = gql.gql(
    """
        query consumptionAndRate($accountNumber:String!, $startAt:DateTime!) {
            properties(accountNumber: $accountNumber) {
                electricityMeterPoints {
                    mpan
                    meters {
                        serialNumber
                        consumption(last: 1, grouping:QUARTER, timezone:"UTC", startAt:$startAt) {
                            edges {
                                node {
                                    value
                                    startAt
                                    endAt
                                }
                            }
                        }
                    }
                    agreements {
                        tariff {
                            ... on StandardTariff {
                                displayName
                                unitRate
                            }
                        }
                    }
                }
            }
        }
    """
)



# SPDX-FileCopyrightText: 2021 The octopy-energy Authors
#
# SPDX-License-Identifier: Apache-2.0

import datetime
import json
import unittest

from .. import Consumption


class ConsumptionTest(unittest.TestCase):
    def test_from_rest(self) -> None:
        rest_result_str = """
            {"consumption":0.209,"interval_start":"2021-11-06T23:30:00Z","interval_end":"2021-11-07T00:00:00Z"}
        """

        self.assertEqual(
            Consumption(
                0.209,
                datetime.datetime(2021, 11, 6, 23, 30, 0, tzinfo=datetime.timezone.utc),
                datetime.datetime(2021, 11, 7, 0, 0, 0, tzinfo=datetime.timezone.utc),
            ),
            Consumption.from_rest(json.loads(rest_result_str)),
        )

    def test_from_graphql(self) -> None:
        gql_result_str = """
            {
                "node": {
                    "value": "0.20900000000000000000",
                    "startAt": "2021-11-06T23:30:00+00:00",
                    "endAt": "2021-11-07T00:00:00+00:00"
                }
            }
        """

        self.assertEqual(
            Consumption(
                0.209,
                datetime.datetime(2021, 11, 6, 23, 30, 0, tzinfo=datetime.timezone.utc),
                datetime.datetime(2021, 11, 7, 0, 0, 0, tzinfo=datetime.timezone.utc),
            ),
            Consumption.from_graphql(json.loads(gql_result_str)),
        )



# SPDX-FileCopyrightText: 2021 The octopy-energy Authors
#
# SPDX-License-Identifier: Apache-2.0
"""Sample command that fetches the information needed for Home Assistant integration."""
import asyncio

import click

from .. import graphql


@click.command()
@click.option(
    "--apikey",
    required=True,
    type=str,
    help="The API key to authenticate to Octopus API.",
)
@click.option(
    "--account-number",
    required=True,
    type=str,
    help="The account number to fetch the tariff for.",
)
def main(*args, **kwargs) -> None:
    return asyncio.run(amain(*args, **kwargs))


async def amain(apikey: str, account_number: str) -> None:
    async with graphql.ClientSession(apikey) as session:
        result = await session.query_account_current_electricity_consumption_and_tariff(
            account_number
        )

        for mpan, (consumptions, tariff) in result.items():
            print(f"MPAN {mpan} — Tariff {tariff.display_name} ({tariff.rate} p/kWh)")
            for meter, consumption in consumptions.items():
                print(
                    f"  {meter} reads {consumption.consumption} kWh as of {consumption.interval_end}"
                )


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2021 The octopy-energy Authors
#
# SPDX-License-Identifier: Apache-2.0
"""Simple command line tool to fetch the current electricity tariff on account."""
import asyncio
import csv
import sys

import click

from .. import graphql


@click.command()
@click.option(
    "--apikey",
    required=True,
    type=str,
    help="The API key to authenticate to Octopus API.",
)
@click.option(
    "--account-number",
    required=True,
    type=str,
    help="The account number to fetch the tariff for.",
)
def main(*args, **kwargs) -> None:
    return asyncio.run(amain(*args, **kwargs))


async def amain(apikey: str, account_number: str) -> None:
    writer = csv.writer(sys.stdout)
    writer.writerow(["MPAN", "Tariff Name", "Unit Rate"])
    async with graphql.ClientSession(apikey) as session:
        result = await session.query_account_active_electricity_tariffs(account_number)
        for mpan, tariff in result.items():
            writer.writerow([mpan, tariff.display_name, tariff.rate])


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2021 The octopy-energy Authors
#
# SPDX-License-Identifier: Apache-2.0

import unittest

from .. import ConsumptionGroupings


class GroupingsRestTest(unittest.TestCase):
    """Test the conversion of groupings to REST parameters."""

    def test_conversion_halfhour(self) -> None:
        """The half-hour grouping is the default, and does not have a matching value in REST."""
        self.assertEqual("", ConsumptionGroupings.HALF_HOUR.rest)

    def test_conversion_hour(self) -> None:
        self.assertEqual("hour", ConsumptionGroupings.HOUR.rest)


class GroupingsGraphQLTest(unittest.TestCase):
    """Test the conversion of groupings to GraphQL parameters."""

    def test_conversion_halfhour(self) -> None:
        self.assertEqual("HALF_HOUR", ConsumptionGroupings.HALF_HOUR.graphql)

    def test_conversion_hour(self) -> None:
        self.assertEqual("HOUR", ConsumptionGroupings.HOUR.graphql)



# SPDX-FileCopyrightText: 2021 The octopy-energy Authors
#
# SPDX-License-Identifier: Apache-2.0
"""Common data types for Octopus Energy client."""

import dataclasses
import datetime
import enum
from typing import Any

import iso8601


@enum.unique
class ConsumptionGroupings(enum.Enum):
    HALF_HOUR = ""
    HOUR = "hour"
    DAY = "day"
    WEEK = "week"
    MONTH = "month"
    QUARTER = "quarter"

    @property
    def rest(self) -> str:
        return self.value

    @property
    def graphql(self) -> str:
        return self.name


@dataclasses.dataclass
class Consumption:
    consumption: float
    interval_start: datetime.datetime
    interval_end: datetime.datetime

    @classmethod
    def from_rest(cls, json_dict: dict[str, Any]) -> "Consumption":
        return Consumption(
            json_dict["consumption"],
            iso8601.parse_date(json_dict["interval_start"]),
            iso8601.parse_date(json_dict["interval_end"]),
        )

    @classmethod
    def from_graphql(cls, gql_dict: dict[str, Any]) -> "Consumption":
        if "node" in gql_dict:
            return cls.from_graphql(gql_dict["node"])

        return Consumption(
            float(gql_dict["value"]),
            iso8601.parse_date(gql_dict["startAt"]),
            iso8601.parse_date(gql_dict["endAt"]),
        )


@dataclasses.dataclass
class Tariff:
    display_name: str
    rate: float

    @classmethod
    def from_graphql(cls, gql_dict: dict[str, Any]) -> "Tariff":
        if "tariff" in gql_dict:
            return cls.from_graphql(gql_dict["tariff"])

        return Tariff(gql_dict["displayName"], gql_dict["unitRate"])



# -*- coding: utf-8 -*-
#
# SPDX-FileCopyrightText: 2021 The octopy-energy Authors
# SPDX-License-Identifier: 0BSD

# Ensure it's present.
import setuptools_scm  # noqa: F401
from setuptools import setup

setup()



# SPDX-FileCopyrightText: 2021 The octopy-energy Authors
#
# SPDX-License-Identifier: Apache-2.0
"""Simple command line tool to fetch the electricity readings."""
import asyncio

import click

from .. import rest


@click.command()
@click.option(
    "--apikey",
    required=True,
    type=str,
    help="The API key to authenticate to Octopus API.",
)
@click.option(
    "--mpan",
    required=True,
    type=str,
    help="The MPAN number of the electricity meter point to query.",
)
@click.option(
    "--meter", required=True, type=str, help="The meter serial number to query."
)
def main(*args, **kwargs) -> None:
    return asyncio.run(amain(*args, **kwargs))


async def amain(apikey: str, mpan: str, meter: str) -> None:
    async with rest.ClientSession(apikey=apikey) as session:
        async for reading in session.electricity_meter_consumption(mpan, meter):
            print(reading)


if __name__ == "__main__":
    main()



# SPDX-FileCopyrightText: 2021 The octopy-energy Authors
#
# SPDX-License-Identifier: Apache-2.0
"""REST client for Octopus Energy APIs."""

import datetime
from collections.abc import Iterator
from typing import AsyncIterator, Optional, Union, cast

import aiohttp
import yarl

from . import Consumption, ConsumptionGroupings

_OCTOPUS_ENERGY_BASE_URL = "https://api.octopus.energy/"


class ClientSession(aiohttp.ClientSession):
    """Façade for aiohttp.ClientSession supporting Octopus Energy API."""

    _base_api_url: yarl.URL

    def __init__(self, apikey: str, *args, **kwargs) -> None:
        if "auth" in kwargs:
            raise ValueError("Invalid argument 'auth', please only provide 'apikey'.")
        if not kwargs.pop("raise_for_status", True):
            raise ValueError("Invalid argument 'raise_for_status', please remove.")

        # We save this separately because some of the APIs use fully qualified URLs, so
        # we cannot let aiohttp know about the base_url!
        self._base_api_url = yarl.URL(kwargs.pop("base_url", _OCTOPUS_ENERGY_BASE_URL))

        super().__init__(
            *args, auth=aiohttp.BasicAuth(apikey), raise_for_status=True, **kwargs
        )

    # This is because mypy complains about the additional methods otherwise.
    async def __aenter__(self):
        return cast("ClientSession", await super().__aenter__())

    def api_url(self, api_url: Union[str, yarl.URL]) -> yarl.URL:
        api_url = yarl.URL(api_url)

        if not api_url.is_absolute():
            return self._base_api_url.join(api_url)

        return api_url

    def electricity_meter_consumption(
        self, *args, **kwargs
    ) -> AsyncIterator[Consumption]:
        return ElectricityMeterConsumption(self, *args, **kwargs)


class ElectricityMeterConsumption:

    _results: Iterator[Consumption] = iter(())
    _next_url: Optional[yarl.URL] = None

    def __init__(
        self,
        session: ClientSession,
        mpan: str,
        serial_number: str,
        period_range: Union[None, tuple[datetime.datetime, datetime.datetime]] = None,
        grouping: ConsumptionGroupings = ConsumptionGroupings.HALF_HOUR,
    ) -> None:
        self._session = session

        self._next_url = session.api_url(
            f"/v1/electricity-meter-points/{mpan}/meters/{serial_number}/consumption"
        )

        params = {}
        if period_range:
            period_from, period_to = period_range
            params["period_from"] = period_from.isoformat()
            params["period_to"] = period_to.isoformat()

        # Half-hour grouping requires *not* passing a parameter at all.
        if grouping.rest:
            params["group_by"] = grouping.rest

        if params:
            self._next_url = self._next_url.with_query(**params)

    def __aiter__(self) -> AsyncIterator[Consumption]:
        return self

    async def __anext__(self) -> Consumption:
        try:
            return next(self._results)
        except StopIteration:
            if self._next_url is None:
                raise StopAsyncIteration()

            async with self._session.get(self._next_url) as response:
                response_dict = await response.json()

            self._next_url = response_dict["next"]
            self._results = (
                Consumption.from_rest(result) for result in response_dict["results"]
            )

            return await self.__anext__()
